"""
This module implements method to generate a deterministic, physics-inspired
toy dataset. The dataset is intended for documentations and examples. The
module does not rely on external random number generators (seeding numpy
might break user code).
"""

import math

import numpy as np
from scipy.integrate import quad
from scipy.interpolate import interp1d
from numba import njit, jit
import pandas as pd

from pylorentz import Momentum4

import warnings
warnings.simplefilter("error")

def rand(size=1, seed=1991, return_final=False):
    """
    Returns a numpy array with random floats between 0 and 1. Calling the
    method twice (with the same seed) returns the same array.

    The parameter n defines the length of the returned array.

    If return_final is true, the method returns the final integer value
    which can be used as a seed for the next iteration.
    """
    # This method uses a Linear congruential generator with the same
    # parameters used in Virtual Pascal.
    # https://en.wikipedia.org/wiki/Linear_congruential_generator
    m = 2**32
    a = 134775813
    c = 1 

    output = np.empty(size, dtype='int')
    seed = (a * seed + c) % m  # make the output look more random
    output[0] = (a * seed + c) % m

    for i in range(1, size):
        output[i] = (a * output[i - 1] + c) % m

    if return_final:
        return output / m, output[-1]

    return output / m

class Random:
    """
    State based random generator.
    """
    def __init__(self, seed=1991):
        """
        Creates and seeds an independent random generator.
        """
        self.seed = seed

    def rand(self, size=1):
        """
        Returns a numpy array w
        """
        output, self.seed = rand(size, self.seed, return_final=True)
        return output


def pdf_gluon(x):
    """
    Grossly simplification of the parton distribution function of a proton for
    gluons. The method returns the probability density to find a gluon within
    the proton carrying a momentum fraction x.

    The pdf vanishes for x = 1.
    >>> '%.4f' % pdf_gluon(1)
    '0.0000'

    The pdf is normalized to unity.
    """
    return 7 * (x - 1)**6

def pdf_quark(x):
    """
    Grossly simplification of the parton distribution function of a proton for
    gluons. The method returns the probability density to find a gluon within
    the proton carrying a momentum fraction x.

    The quark pdf has a bump at x=0.5.

    The pdf almost vanishes for x = 1.
    >>> pdf_quark(1) < 0.1
    True

    The pdf is normalized to unity.
    """
    return 0.6 * 7 * (x - 1)**6 \
           + 0.4 / (20 * math.atan(5) * ((x - 0.5)**2 + 0.01))

def ipdf(pdf, x, dx):
    """
    Integrate the given pdf (callable) from x to x + dx.

    >>> '%.4f' % ipdf(pdf_gluon, 0, 1)
    '1.0000'
    >>> '%.4f' % ipdf(pdf_quark, 0, 1)
    '1.0000'
    """
    return quad(pdf, x, x + dx)[0]

def draw(pdf, rng=None, *args, **kwds):
    """
    Draws a random x from the given pdf. All arguments are forwarded to
    rand().
    """
    N = 100
    grid = np.linspace(0, 1, N)
    cdf = np.array([quad(pdf, 0, x)[0] for x in grid])

    inv_cdf = interp1d(cdf, grid, bounds_error=False,
                       fill_value='extrapolate')

    if rng is None:
        ys = rand(*args, **kwds)
    else:
        ys = rng.rand(*args, **kwds)

    return inv_cdf(ys)

def gaussp_pdf(x):
    """
    Probability density function of a Gaussian centered around 0.5 with a
    stddev of 0.1.
    """

    return np.exp(-0.5 * (x - 0.5)**2 / 0.01) / 0.25066268375731304

def thetap_pdf(x):
    """
    Probability density function of the transformed theta angel (scalted to
    match [0, 1)) if every point on the sphere is equally likely.
    """
    return math.pi * np.sin(math.pi * x) / 2

def random_direction(rng, *args, **kwds):
    """
    Returns a random direction parameterized by eta and phi. The first
    argument must be a Random object. All other arguments are passed on to
    draw() and Random.rand().

    Return eta, phi.
    """
    phi = (rng.rand(*args, **kwds) - 0.5) * 2 * math.pi
    thetap = draw(thetap_pdf, rng=rng, *args, **kwds)
    theta = math.pi * thetap
    a = theta / 2
    b = np.tan(a)
    eta = - np.log(b)

    return eta, phi

def reverse(p):
    """
    Multiplies the space components by -1.
    """
    return p.__class__(p[0], -p[1], -p[2], -p[3])

def build_vbfh_tautau(rng, size=10):
    """
    Builds a vector-boson fusion event producing a Higgs boson which decays to
    a pair of tau leptons.

    This has nothing to do with real physics!
    """
    data = {
        "mu_1_pt": [],
        "mu_1_eta": [],
        "mu_1_phi": [],
        "mu_2_pt": [],
        "mu_2_eta": [],
        "mu_2_phi": [],
        "met_pt": [],
        "met_phi": [],
        "jet_1_pt": [],
        "jet_1_eta": [],
        "jet_1_phi": [],
        "jet_2_pt": [],
        "jet_2_eta": [],
        "jet_2_phi": [],
        "signal": [],
        "m_vis": [],
    }
            
    E_beam = 4500
    for _ in range(size):
        q_1 = draw(pdf_quark, rng)[0] * E_beam
        q_1 = Momentum4(math.sqrt(q_1**2 + 1), 0, 0, q_1)

        q_2 = draw(pdf_quark, rng)[0] * E_beam
        q_2 = Momentum4(math.sqrt(q_2**2 + 1), 0, 0, -q_2)

        # Vector-bosons have some fraction of quark energy in a random direction
        v_1 = q_1.e * 0.2
        v_1 = Momentum4.e_m_eta_phi(math.sqrt(90**2 + v_1**2),
                                    90,
                                    *[x[0] for x in random_direction(rng)])
        v_2 = q_2.e * 0.2
        v_2 = Momentum4.e_m_eta_phi(math.sqrt(90**2 + v_2**2),
                                    90,
                                    *[x[0] for x in random_direction(rng)])

        v_sum = v_1 + v_2
        smear = draw(gaussp_pdf, rng)[0] + 0.5
        higgs = Momentum4.e_m_eta_phi(v_sum.e, 125 * smear, v_sum.eta, v_sum.phi)

        q_1_final = q_1 - higgs / 2
        q_2_final = q_2 - higgs / 2

        if abs(q_1_final.eta) > abs(q_2_final.eta):
            q_1_final, q_2_final = q_2_final, q_1_final

        gluon = q_1.e * 0.02
        gluon = Momentum4.e_m_eta_phi(math.sqrt(gluon**2 + 1),
                                      1,
                                      *[x[0] for x in random_direction(rng)])
        gluon = gluon.boost_particle(q_1)
        q_1_final = q_1_final - gluon

        # tau 1
        tau_1 = Momentum4.e_m_eta_phi(higgs.m / 2, 1.777,
                                      *[x[0] for x in random_direction(rng)])
        tau_1 = tau_1.boost_particle(higgs)
        nu_tau_1 = Momentum4.e_m_eta_phi(tau_1.m / 2, 0,
                                         *[x[0] for x in random_direction(rng)])
        nu_mu_1 = Momentum4.e_m_eta_phi(tau_1.m / 4, 0.1,
                                        *[x[0] for x in random_direction(rng)])
        mu_1 = reverse(nu_mu_1)

        nu_tau_1 = nu_tau_1.boost_particle(tau_1)
        nu_mu_1 = nu_mu_1.boost_particle(tau_1)
        mu_1 = mu_1.boost_particle(tau_1)

        # tau 2
        tau_2 = reverse(tau_1)
        tau_2 = tau_2.boost_particle(higgs)
        nu_tau_2 = Momentum4.e_m_eta_phi(tau_2.m / 2, 0,
                                         *[x[0] for x in random_direction(rng)])
        nu_mu_2 = Momentum4.e_m_eta_phi(tau_2.m / 4, 0.1,
                                        *[x[0] for x in random_direction(rng)])
        mu_2 = reverse(nu_mu_2)

        nu_tau_2 = nu_tau_2.boost_particle(tau_2)
        nu_mu_2 = nu_mu_2.boost_particle(tau_2)
        mu_2 = mu_2.boost_particle(tau_2)

        met = nu_mu_1 + nu_mu_2 + nu_tau_1 + nu_tau_2

        if mu_1.p_t < mu_2.p_t:
            mu_1, mu_2 = mu_2, mu_1

        jet_1, jet_2 = sorted([q_1_final, q_2_final],
                                     key=lambda p: -p.p_t)

        data["mu_1_pt"].append(mu_1.p_t)
        data["mu_1_eta"].append(mu_1.eta)
        data["mu_1_phi"].append(mu_1.phi)
        data["mu_2_pt"].append(mu_2.p_t)
        data["mu_2_eta"].append(mu_2.eta)
        data["mu_2_phi"].append(mu_2.phi)
        data["met_pt"].append(met.p_t)
        data["met_phi"].append(met.phi)
        data["jet_1_pt"].append(jet_1.p_t)
        data["jet_1_eta"].append(jet_1.eta)
        data["jet_1_phi"].append(jet_1.phi)
        data["jet_2_pt"].append(jet_2.p_t)
        data["jet_2_eta"].append(jet_2.eta)
        data["jet_2_phi"].append(jet_2.phi)
        data["signal"].append(True)
        data["m_vis"].append((mu_1 + mu_2).m)

    return pd.DataFrame(data)
